@@ -1,23 +1,27 @@
 /*
  * Author: Copyright (C) Andrzej Surowiec 2012
  *                      Parts Rudolf Boeddeker  Date: 2013-08-13
+ * Copyright (c) 2012-2018 Nitrokey UG
  *
- * This file is part of Nitrokey.
+ * This file is part of Nitrokey App.
  *
- * Nitrokey is free software: you can redistribute it and/or modify
+ * Nitrokey App is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * any later version.
  *
- * Nitrokey is distributed in the hope that it will be useful,
+ * Nitrokey App is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with Nitrokey. If not, see <http://www.gnu.org/licenses/>.
+ * along with Nitrokey App. If not, see <http://www.gnu.org/licenses/>.
+ *
+ * SPDX-License-Identifier: GPL-3.0
  */
-#include "libnitrokey/include/NitrokeyManager.h"
+
+#include <libnitrokey/NitrokeyManager.h>
 
 #include "mainwindow.h"
 #include "aboutdialog.h"
@@ -43,21 +47,98 @@
 #include "src/core/SecureString.h"
 
 #include <QString>
-#include <src/core/ScopedGuard.h>
+#include "src/core/ScopedGuard.h"
+#include "src/core/ThreadWorker.h"
+#include "hotpslot.h"
+
+#include <QFuture>
+#include <QtConcurrent/QtConcurrent>
+#include <QSvgWidget>
 
 using nm = nitrokey::NitrokeyManager;
-static const QString communication_error_message = QApplication::tr("Communication error. Please reinsert the device.");
+const auto Communication_error_message = QT_TRANSLATE_NOOP("MainWindow", "Communication error. Please reinsert the device.");
 
 
-static const QString invalid_secret_key_string_details = QApplication::tr("The secret string you have entered is invalid. Please reenter it.")
-                           +"\n"+QApplication::tr("Details: ");
+const auto Invalid_secret_key_string_details_1 = QT_TRANSLATE_NOOP("MainWindow", "The secret string you have entered is invalid. Please reenter it.");
+const auto Invalid_secret_key_string_details_2 = /*"\n" +*/ QT_TRANSLATE_NOOP("MainWindow", "Details: ");
+
+const auto Warning_ev_not_secure_initialize = QT_TRANSLATE_NOOP("MainWindow", "Warning: Encrypted volume is not secure,\nSelect \"Initialize "
+                                                           "device\" option from context menu.");
+
+
+const auto Reset_nitrokeys_time = QT_TRANSLATE_NOOP("MainWindow", "Reset Nitrokey's time?");
+const auto Warning_devices_clock_not_desynchronized =
+    QT_TRANSLATE_NOOP("MainWindow", "WARNING!\n\nThe time of your computer and Nitrokey are out of "
+                         "sync. Your computer may be configured with a wrong time or "
+                         "your Nitrokey may have been attacked. If an attacker or "
+                         "malware could have used your Nitrokey you should reset the "
+                         "secrets of your configured One Time Passwords. If your "
+                         "computer's time is wrong, please configure it correctly and "
+                         "reset the time of your Nitrokey.\n\nReset Nitrokey's time?");
+
+const auto Tray_location_msg = /*"\n" +*/ QT_TRANSLATE_NOOP("MainWindow", "You can find application’s tray icon in system tray in "
+                                      "the right down corner of your screen (Windows) or in the upper right (Linux, MacOS).");
+
+void MainWindow::load_settings_page(){
+    QSettings settings;
+    const auto first_run_key = "main/first_run";
+    auto first_run = settings.value(first_run_key, true).toBool();
+    ui->cb_first_run_message->setChecked(first_run);
+
+    const auto lang_key = "main/language";
+    auto lang_selected = settings.value(lang_key, "-----").toString();
+
+    ui->combo_languages->clear();
+    QDir langDir(":/i18n/");
+    auto list = langDir.entryList();
+    for (auto &&translationFile : list) {
+      auto lang = translationFile.remove("nitrokey_").remove(".qm");
+      ui->combo_languages->addItem(lang, lang);
+    }
+    ui->combo_languages->addItem(tr("current:") +" "+ lang_selected, lang_selected);
+    ui->combo_languages->setCurrentText(tr("current:") +" "+ lang_selected);
+
+    ui->edit_debug_file_path->setText(settings.value("debug/file", "").toString());
+    ui->spin_debug_verbosity->setValue(settings.value("debug/level", 2).toInt());
+    auto settings_debug_enabled = settings.value("debug/enabled", false).toBool();
+    ui->cb_debug_enabled->setChecked(settings_debug_enabled);
+    emit ui->cb_debug_enabled->toggled(settings_debug_enabled);
+    ui->spin_PWS_time->setValue(settings.value("clipboard/PWS_time", 60).toInt());
+    ui->spin_OTP_time->setValue(settings.value("clipboard/OTP_time", 120).toInt());
+    ui->cb_device_connection_message->setChecked(settings.value("main/connection_message", true).toBool());
+    ui->cb_show_main_window_on_connection->setChecked(settings.value("main/show_main_on_connection", true).toBool());
+    ui->cb_hide_main_window_on_connection->setChecked(settings.value("main/close_main_on_connection", false).toBool());
+    ui->cb_hide_main_window_on_close->setChecked(settings.value("main/hide_on_close", true).toBool());
+
+  ui->cb_check_symlink->setChecked(settings.value("storage/check_symlink", false).toBool());
+#ifndef Q_OS_LINUX
+    ui->cb_check_symlink->setEnabled(false);
+#endif
+}
+
+void MainWindow::keyPressEvent(QKeyEvent *keyevent)
+{
+    if (keyevent->key()==Qt::Key_Escape){
+        QSettings settings;
+        if (settings.value("main/hide_on_close", true).toBool()){
+#ifdef Q_OS_MAC
+            showMinimized();
+#else
+            hide();
+#endif
+        } else
+            QApplication::quit();
+    } else {
+        QMainWindow::keyPressEvent(keyevent);
+    }
+}
 
 MainWindow::MainWindow(QWidget *parent):
     QMainWindow(parent),
     ui(new Ui::MainWindow),
     clipboard(this),
-    auth_admin(this, Authentication::Type::ADMIN),
-    auth_user(this, Authentication::Type::USER),
+    auth_admin(nullptr, Authentication::Type::ADMIN),
+    auth_user(nullptr, Authentication::Type::USER),
     storage(this, &auth_admin, &auth_user),
     tray(this, false, false, &storage),
     HOTP_SlotCount(HOTP_SLOT_COUNT),
@@ -78,6 +159,8 @@ MainWindow::MainWindow(QWidget *parent):
   connect(&storage, SIGNAL(FactoryReset()), &tray, SLOT(regenerateMenu()));
   connect(&storage, SIGNAL(FactoryReset()), libada::i().get(), SLOT(on_FactoryReset()));
   connect(&storage, SIGNAL(longOperationStarted()), this, SLOT(on_KeepDeviceOnline()));
+  connect(&storage, SIGNAL(longOperationStarted()), this, SLOT(on_longOperationStart()));
+
   connect(this, SIGNAL(FactoryReset()), &tray, SLOT(regenerateMenu()));
   connect(this, SIGNAL(FactoryReset()), libada::i().get(), SLOT(on_FactoryReset()));
 
@@ -85,6 +168,7 @@ MainWindow::MainWindow(QWidget *parent):
   connect(this, SIGNAL(PWS_unlocked()), this, SLOT(SetupPasswordSafeConfig()));
   connect(this, SIGNAL(DeviceLocked()), this, SLOT(SetupPasswordSafeConfig()));
   connect(&storage, SIGNAL(storageStatusChanged()), this, SLOT(SetupPasswordSafeConfig()));
+  connect(&storage, SIGNAL(storageStatusChanged()), this, SLOT(storage_check_symlink()));
   connect(this, SIGNAL(PWS_slot_saved(int)), &tray, SLOT(regenerateMenu()));
   connect(this, SIGNAL(PWS_slot_saved(int)), libada::i().get(), SLOT(on_PWS_save(int)));
 
@@ -107,6 +191,13 @@ MainWindow::MainWindow(QWidget *parent):
   connect(this, SIGNAL(DeviceConnected()), &auth_user, SLOT(clearTemporaryPasswordForced()));
   connect(this, SIGNAL(DeviceLocked()), &auth_admin, SLOT(clearTemporaryPasswordForced()));
   connect(this, SIGNAL(DeviceLocked()), &auth_user, SLOT(clearTemporaryPasswordForced()));
+  connect(this, SIGNAL(DeviceConnected()), this, SLOT(ready()));
+
+  connect(this, SIGNAL(DeviceConnected()), this, SLOT(manageStartPage()));
+  connect(&storage, SIGNAL(storageStatusUpdated()), this, SLOT(manageStartPage()));
+  connect(&storage, SIGNAL(storageStatusChanged()), this, SLOT(manageStartPage()));
+  connect(this, SIGNAL(PWS_unlocked()), this, SLOT(manageStartPage()));
+  connect(this, SIGNAL(DeviceLocked()), this, SLOT(manageStartPage()));
 
   ui->setupUi(this);
   ui->tabWidget->setCurrentIndex(0); // Set first tab active
@@ -142,7 +233,79 @@ MainWindow::MainWindow(QWidget *parent):
   this->adjustSize();
   this->move(QApplication::desktop()->screen()->rect().center() - this->rect().center());
 
+  tray.setFile_menu(menuBar());
+  ui->progressBar->hide();
+
   first_run();
+  load_settings_page();
+
+  make_UI_enabled(false);
+  startConfiguration(false);
+
+  ui->statusBar->showMessage(tr("Idle"));
+
+  // Connect dial page
+  connect(ui->btn_dial_PWS, SIGNAL(clicked()), this, SLOT(PWS_Clicked_EnablePWSAccess()));
+  connect(ui->btn_dial_lock, SIGNAL(clicked()), this, SLOT(startLockDeviceAction()));  
+  connect(ui->btn_dial_EV, SIGNAL(clicked()), &storage, SLOT(startStick20EnableCryptedVolume()));
+  connect(ui->btn_dial_help, SIGNAL(clicked()), this, SLOT(startHelpAction()));
+  connect(ui->btn_dial_HV, SIGNAL(clicked()), &storage, SLOT(startStick20EnableHiddenVolume()));
+  connect(ui->btn_dial_quit, SIGNAL(clicked()), qApp, SLOT(quit()));
+
+
+}
+
+#include <libnitrokey/stick20_commands.h>
+
+void MainWindow::make_UI_enabled(bool enabled) {
+  ui->tab->setEnabled(enabled);
+  ui->tab_2->setEnabled(enabled);
+  ui->tab_3->setEnabled(enabled);
+  if(!enabled){
+    ui->btn_dial_PWS->setEnabled(enabled);
+    ui->btn_dial_lock->setEnabled(enabled);
+    ui->btn_dial_EV->setEnabled(enabled);
+    ui->btn_dial_HV->setEnabled(enabled);
+  }
+}
+
+void MainWindow::manageStartPage(){
+    if (!libada::i()->isDeviceConnected() || libada::i()->get_status_no_except() !=0 ) {
+      make_UI_enabled(false);
+      return;
+    }
+    make_UI_enabled(true);
+
+
+    QFuture<bool> PWS_unlocked = QtConcurrent::run(libada::i().get(), &libada::isPasswordSafeUnlocked);
+    PWS_unlocked.waitForFinished();
+    ui->btn_dial_PWS->setEnabled(!PWS_unlocked.result());
+    ui->btn_dial_lock->setEnabled(PWS_unlocked.result());
+
+    auto is_storage = libada::i()->isStorageDeviceConnected();
+    ui->btn_dial_EV->setEnabled(false);
+    ui->btn_dial_HV->setEnabled(false);
+
+    if (is_storage){
+        QFuture<nitrokey::stick20::DeviceConfigurationResponsePacket::ResponsePayload> status
+                = QtConcurrent::run(nitrokey::NitrokeyManager::instance().get(),
+                                    &nitrokey::NitrokeyManager::get_status_storage);
+        status.waitForFinished();
+
+        bool initialized = !status.result().StickKeysNotInitiated && status.result().SDFillWithRandomChars_u8;
+        if (!initialized){
+          make_UI_enabled(false);
+        }
+
+        ui->btn_dial_PWS->setEnabled(initialized && !PWS_unlocked.result());
+        ui->btn_dial_EV->setEnabled(initialized && !status.result().VolumeActiceFlag_st.encrypted);
+        ui->btn_dial_HV->setEnabled(initialized && !status.result().VolumeActiceFlag_st.hidden);
+        ui->btn_dial_lock->setEnabled(status.result().VolumeActiceFlag_st.encrypted
+                                      || status.result().VolumeActiceFlag_st.hidden
+                                      || PWS_unlocked.result()
+                                      );
+        ui->PWS_ButtonEnable->setEnabled(ui->btn_dial_PWS->isEnabled());
+    }
 }
 
 void MainWindow::first_run(){
@@ -150,12 +313,15 @@ void MainWindow::first_run(){
   const auto first_run_key = "main/first_run";
   auto first_run = settings.value(first_run_key, true).toBool();
   if (!first_run) return;
-  settings.setValue(first_run_key, false);
 
-  auto msg = tr(
-        "The Nitrokey App is available as an icon in the tray bar."
-  );
-  csApplet()->messageBox(msg);
+  auto msg = tr("The Nitrokey App is available as an icon in the tray bar.");
+  tray.showTrayMessage(msg);
+  msg += "\n" + tr(Tray_location_msg);
+  msg += " " + tr("Would you like to show this message again?");
+  bool user_wants_to_be_reminded = csApplet()->yesOrNoBox(msg, true);
+  if(!user_wants_to_be_reminded){
+      settings.setValue(first_run_key, false);
+  }
 
   //TODO insert call to First run configuration wizard here
 }
@@ -213,37 +379,42 @@ void MainWindow::checkConnection() {
 }
 
 void MainWindow::initialTimeReset() {
-  if (!libada::i()->isDeviceConnected()) {
+  if (!libada::i()->isDeviceConnected() || long_operation_in_progress) {
     return;
   }
 
-  if (!libada::i()->is_time_synchronized()) {
-    bool answer = csApplet()->detailedYesOrNoBox(tr("Time is out-of-sync"),
-      tr("WARNING!\n\nThe time of your computer and Nitrokey are out of "
-                 "sync. Your computer may be configured with a wrong time or "
-                 "your Nitrokey may have been attacked. If an attacker or "
-                 "malware could have used your Nitrokey you should reset the "
-                 "secrets of your configured One Time Passwords. If your "
-                 "computer's time is wrong, please configure it correctly and "
-                 "reset the time of your Nitrokey.\n\nReset Nitrokey's time?"),
+  QFuture<bool> is_time_synchronized = QtConcurrent::run(libada::i().get(), &libada::is_time_synchronized);
+  is_time_synchronized.waitForFinished();
+
+  if (!is_time_synchronized.result()) {
+    bool answer = csApplet()->detailedYesOrNoBox(tr("Time is out-of-sync") + " - " + tr(Reset_nitrokeys_time), tr(Warning_devices_clock_not_desynchronized),
       false);
     if (answer) {
       auto res = libada::i()->set_current_time();
       if (res) {
-        csApplet()->messageBox(tr("Time reset!"));
+          tray.showTrayMessage(tr("Time reset!"));
       }
     }
   }
 }
 
 MainWindow::~MainWindow() {
-  nm::instance()->set_log_function([](std::string data){});
+  nm::instance()->set_log_function([](std::string /*data*/){});
   delete ui;
 }
 
 void MainWindow::closeEvent(QCloseEvent *event) {
-  this->hide();
-  event->ignore();
+  QSettings settings;
+  if (settings.value("main/hide_on_close", true).toBool()){
+#ifdef Q_OS_MAC
+     showMinimized();
+#else
+      hide();
+#endif
+      event->ignore();
+  } else {
+      QApplication::quit();
+  }
 }
 
 void MainWindow::generateComboBoxEntrys() {
@@ -526,22 +697,31 @@ void MainWindow::displayCurrentGeneralConfig() {
   ui->enableUserPasswordCheckBox->setChecked(status.enable_user_password != 0);
   ui->deleteUserPasswordCheckBox->setChecked(status.delete_user_password != 0);
 }
+void MainWindow::startConfigurationMain() {
+    startConfiguration(false);
+    ui->tabWidget->setCurrentIndex(0);
+}
+
+void MainWindow::startConfiguration(bool changeTab) {
+    if (long_operation_in_progress) return;
 
-void MainWindow::startConfiguration() {
-  bool validPassword = true;
-  // Start the config dialog
-  if (validPassword) {
     displayCurrentSlotConfig();
     displayCurrentGeneralConfig();
     SetupPasswordSafeConfig();
 
-    ManageWindow::bringToFocus(this);
+    if (libada::i()->isStorageDeviceConnected()) {
+      ui->counterEdit->setMaxLength(7);
+    }
 
     QTimer::singleShot(0, this, SLOT(resizeMin()));
-  }
-  if (libada::i()->isStorageDeviceConnected()) {
-    ui->counterEdit->setMaxLength(7);
-  }
+
+    if (changeTab){
+      ui->tabWidget->setCurrentIndex(1);
+    }
+    QTimer::singleShot(0, this, [this](){
+      ManageWindow::bringToFocus(this);
+      make_UI_enabled(libada::i()->isDeviceConnected());
+    });
 }
 
 void MainWindow::resizeMin() { resize(minimumSizeHint()); }
@@ -552,6 +732,18 @@ void MainWindow::startStickDebug() {
   debug->raise();
 }
 
+#include <QDesktopServices>
+#include <QUrl>
+void MainWindow::startHelpAction() {
+    QString link = "https://www.nitrokey.com/start";
+    QDesktopServices::openUrl(QUrl(link));
+}
+
+void MainWindow::on_longOperationStart(){
+  hide();
+  emit LongOperationStart();
+}
+
 void MainWindow::startAboutDialog() {
   AboutDialog dialog(this);
   dialog.exec();
@@ -586,10 +778,21 @@ void MainWindow::startResetUserPassword() {
 
 
 void MainWindow::storage_check_symlink(){
-    if (!QFileInfo("/dev/nitrospace").isSymLink()) {
-        csApplet()->warningBox(tr("Warning: The encrypted Volume is not formatted.\n\"Use GParted "
-                                          "or fdisk for this.\""));
+//TODO make better partition detection method
+#ifdef Q_OS_LINUX
+    QSettings settings;
+    bool should_remind = settings.value("storage/check_symlink", true).toBool();
+    if (should_remind && !QFileInfo("/dev/nitrospace").isSymLink()) {
+        bool user_wants_reminding = csApplet()->yesOrNoBox(
+                                    tr("Warning: Application could not detect any partition on the Encrypted Volume. "
+                                       "Please use graphical GParted or terminal fdisk/parted tools for this.")+
+                                       " " + tr("Would you like to be reminded again?")
+                                    , true);
+        if (!user_wants_reminding){
+            settings.setValue("storage/check_symlink", false);
+        }
     }
+#endif
 }
 
 //int MainWindow::stick20SendCommand(uint8_t stick20Command, uint8_t *password) {
@@ -627,12 +830,12 @@ void MainWindow::on_writeButton_clicked() {
     }
   }
   catch(const cppcodec::parse_error &e){
-    csApplet()->warningBox(invalid_secret_key_string_details + e.what());
+    csApplet()->warningBox(tr(Invalid_secret_key_string_details_1) + "\n" + tr(Invalid_secret_key_string_details_2) + e.what());
     return;
   }
 
   if (!this->ui->secretEdit->text().isEmpty() && !validate_raw_secret(otp.secret)) {
-    csApplet()->warningBox(invalid_secret_key_string_details + tr("secret is not passing validation."));
+    csApplet()->warningBox(tr(Invalid_secret_key_string_details_1) + "\n" + tr(Invalid_secret_key_string_details_2) + tr("secret is not passing validation."));
     return;
   }
 
@@ -690,12 +893,12 @@ void MainWindow::on_hexRadioButton_toggled(bool checked) {
       auto secret_raw = decodeBase32Secret(secret, debug_mode);
       auto secret_hex = QString::fromStdString(cppcodec::hex_upper::encode(secret_raw));
       ui->secretEdit->setText(secret_hex);
-      clipboard.copyToClipboard(secret_hex);
+      clipboard.copyOTP(secret_hex);
       showNotificationLabel();
     }
     catch (const cppcodec::parse_error &e) {
       ui->secretEdit->setText("");
-      csApplet()->warningBox(invalid_secret_key_string_details + e.what());
+      csApplet()->warningBox(tr(Invalid_secret_key_string_details_1) + "\n" + tr(Invalid_secret_key_string_details_2) + e.what());
     }
   }
 }
@@ -735,7 +938,7 @@ void MainWindow::on_base32RadioButton_toggled(bool checked) {
       auto secret_raw = cppcodec::hex_upper::decode(secret_hex);
       auto secret_base32 = QString::fromStdString(base32::encode(secret_raw));
       ui->secretEdit->setText(secret_base32);
-      clipboard.copyToClipboard(secret_base32);
+      clipboard.copyOTP(secret_base32);
       showNotificationLabel();
     }
     catch (const cppcodec::parse_error &e) {
@@ -778,7 +981,7 @@ void MainWindow::on_enableUserPasswordCheckBox_toggled(bool checked) {
 
 void MainWindow::on_writeGeneralConfigButton_clicked() {
   if (!libada::i()->isDeviceConnected()) {
-      csApplet()->warningBox(tr("Nitrokey not connected!"));
+      csApplet()->warningBox(tr("Nitrokey is not connected!"));
   }
     if(!auth_admin.authenticate()){
       csApplet()->warningBox(tr("Wrong PIN. Please try again."));
@@ -809,7 +1012,7 @@ void MainWindow::getHOTPDialog(int slot) {
   try{
     auto OTPcode = getNextCode(0x10 + slot);
     if (OTPcode.empty()) return;
-    clipboard.copyToClipboard(QString::fromStdString(OTPcode));
+    clipboard.copyOTP(QString::fromStdString(OTPcode));
 
 
     if (libada::i()->getHOTPSlotName(slot).empty())
@@ -826,7 +1029,7 @@ void MainWindow::getHOTPDialog(int slot) {
                       TRAY_MSG_TIMEOUT);
   }
   catch(DeviceCommunicationException &e){
-    tray.showTrayMessage(communication_error_message);
+    tray.showTrayMessage(tr(Communication_error_message));
   }
 }
 
@@ -834,7 +1037,7 @@ void MainWindow::getTOTPDialog(int slot) {
   try{
     auto OTPcode = getNextCode(0x20 + slot);
       if (OTPcode.empty()) return;
-    clipboard.copyToClipboard(QString::fromStdString(OTPcode));
+    clipboard.copyOTP(QString::fromStdString(OTPcode));
 
     if (libada::i()->getTOTPSlotName(slot).empty())
       tray.showTrayMessage(QString(tr("TOTP slot ")).append(QString::number(slot + 1, 10)),
@@ -850,7 +1053,7 @@ void MainWindow::getTOTPDialog(int slot) {
                            TRAY_MSG_TIMEOUT);
   }
   catch(DeviceCommunicationException &e){
-    tray.showTrayMessage(communication_error_message);
+    tray.showTrayMessage(tr(Communication_error_message));
   }
 
 }
@@ -910,7 +1113,7 @@ void MainWindow::on_randomSecretButton_clicked() {
   ui->secretEdit->setText(secretArray);
   ui->checkBox->setEnabled(true);
   ui->checkBox->setChecked(true);
-  clipboard.copyToClipboard(secretArray);
+  clipboard.copyOTP(secretArray);
   showNotificationLabel();
   this->checkTextEdited();
 }
@@ -985,6 +1188,7 @@ void MainWindow::checkTextEdited() {
   ui->base32RadioButton->setEnabled(valid);
   ui->hexRadioButton->setEnabled(valid);
   ui->writeButton->setEnabled(valid);
+  ui->btn_copyToClipboard->setEnabled(valid && !secret_key.isEmpty());
 }
 
 void MainWindow::SetupPasswordSafeConfig(void) {
@@ -1028,12 +1232,13 @@ void MainWindow::SetupPasswordSafeConfig(void) {
 
   ui->PWS_ComboBoxSelectSlot->setEnabled(PWS_Access);
   ui->PWS_ButtonEnable->setVisible(!PWS_Access);
+  ui->PWS_Lock->setVisible(PWS_Access);
 
   ui->PWS_EditSlotName->setMaxLength(PWS_SLOTNAME_LENGTH);
   ui->PWS_EditPassword->setMaxLength(PWS_PASSWORD_LENGTH);
   ui->PWS_EditLoginName->setMaxLength(PWS_LOGINNAME_LENGTH);
 
-  ui->PWS_CheckBoxHideSecret->setChecked(TRUE);
+  ui->PWS_CheckBoxHideSecret->setChecked(true);
   ui->PWS_EditPassword->setEchoMode(QLineEdit::Password);
 }
 
@@ -1184,7 +1389,7 @@ void MainWindow::on_PWS_ButtonSaveSlot_clicked() {
     csApplet()->messageBox(tr("Can't save slot. %1").arg(e.last_command_status));
   }
   catch (DeviceCommunicationException &e){
-    csApplet()->messageBox(tr("Can't save slot. %1").arg(communication_error_message));
+    csApplet()->messageBox(tr("Can't save slot. %1").arg(tr(Communication_error_message)));
   }
 }
 
@@ -1203,7 +1408,7 @@ void MainWindow::PWS_Clicked_EnablePWSAccess() {
       return;
     }
     catch (DeviceCommunicationException &e){
-      csApplet()->warningBox(communication_error_message);
+      csApplet()->warningBox(tr(Communication_error_message));
     }
     catch (CommandFailedException &e){
       //TODO emit pw safe not available when not available
@@ -1239,7 +1444,7 @@ void MainWindow::PWS_Clicked_EnablePWSAccess() {
 void MainWindow::PWS_ExceClickedSlot(int Slot) {
   try {
     auto slot_password = nm::instance()->get_password_safe_slot_password((uint8_t) Slot);
-    clipboard.copyToClipboard(slot_password);
+    clipboard.copyPWS(slot_password);
     clear_free_cstr(const_cast<char*>(slot_password));
     QString password_safe_slot_info =
         QString(tr("Password safe [%1]").arg(QString::fromStdString(libada::i()->getPWSSlotName(Slot))));
@@ -1247,7 +1452,7 @@ void MainWindow::PWS_ExceClickedSlot(int Slot) {
     tray.showTrayMessage(title, password_safe_slot_info);
   }
   catch(DeviceCommunicationException){
-    tray.showTrayMessage(communication_error_message);
+    tray.showTrayMessage(tr(Communication_error_message));
   }
 }
 
@@ -1269,19 +1474,12 @@ std::string MainWindow::getNextCode(uint8_t slotNumber) {
     //run only once before first TOTP request
     static bool time_synchronized = libada::i()->is_time_synchronized();
     if (!time_synchronized) {
-       bool user_wants_time_reset = csApplet()->detailedYesOrNoBox(tr("Time is out-of-sync")
-                                                                  + " - " + tr("Reset Nitrokey's time?"),
-                  tr("WARNING!\n\nThe time of your computer and Nitrokey are out of "
-                             "sync.\nYour computer may be configured with a wrong time "
-                             "or\nyour Nitrokey may have been attacked. If an attacker "
-                             "or\nmalware could have used your Nitrokey you should reset the "
-                             "secrets of your configured One Time Passwords. If your "
-                             "computer's time is wrong, please configure it correctly and "
-                             "reset the time of your Nitrokey.\n\nReset Nitrokey's time?"),
-                  false);
+       bool user_wants_time_reset =
+           csApplet()->detailedYesOrNoBox(tr("Time is out-of-sync") + " - " + tr(Reset_nitrokeys_time),
+                                          tr(Warning_devices_clock_not_desynchronized), false);
       if (user_wants_time_reset){
           if(libada::i()->set_current_time()){
-            csApplet()->messageBox(tr("Time reset!"));
+            tray.showTrayMessage(tr("Time reset!"));
             time_synchronized = true;
           }
       }
@@ -1346,10 +1544,10 @@ int MainWindow::factoryResetAction() {
     catch (CommandFailedException &e){
       if(!e.reason_wrong_password())
         throw;
-      csApplet()->messageBox("Wrong Pin. Please try again.");
+      csApplet()->messageBox(tr("Wrong Pin. Please try again."));
       continue;
     }
-    csApplet()->messageBox("Factory reset was successful.");
+    csApplet()->messageBox(tr("Factory reset was successful."));
     emit FactoryReset();
     return 1;
   }
@@ -1422,17 +1620,19 @@ void MainWindow::on_DeviceDisconnected() {
     qDebug("on_DeviceDisconnected");
 
   emit ShortOperationEnds();
-  ui->statusBar->showMessage(tr("Nitrokey disconnected"));
-  tray.showTrayMessage(tr("Nitrokey disconnected"));
 
-  if(this->isVisible()){
-    this->close();
-    csApplet()->messageBox(tr("Closing window due to device disconnection"));
+  QSettings settings;
+  if(settings.value("main/connection_message", true).toBool()){
+    tray.showTrayMessage(tr("Nitrokey disconnected"));
   }
-}
+  ui->statusBar->showMessage(tr("Nitrokey disconnected"));
 
-#include "src/core/ThreadWorker.h"
-#include "hotpslot.h"
+  if(this->isVisible() && settings.value("main/close_main_on_connection", false).toBool()){
+    this->hide();
+  }
+
+  make_UI_enabled(false);
+}
 
 void MainWindow::on_DeviceConnected() {
   if (debug_mode)
@@ -1441,20 +1641,17 @@ void MainWindow::on_DeviceConnected() {
   if (debug_mode)
     emit ShortOperationBegins(tr("Connecting device"));
 
-  //TODO share device state to improve performance
-  try{
-    libada::i()->get_status();
-  }
-  catch (LongOperationInProgressException &e){
-    long_operation_in_progress = true;
-    return;
-  }
+  ui->statusBar->showMessage(tr("Device connected. Waiting for initialization..."));
 
-  auto connected_device_model = libada::i()->isStorageDeviceConnected() ?
-                                tr("Nitrokey Storage connected") :
-                                tr("Nitrokey Pro connected");
-  ui->statusBar->showMessage(connected_device_model);
-  tray.showTrayMessage(tr("Nitrokey connected"), connected_device_model);
+  auto result = QtConcurrent::run(libada::i().get(), &libada::get_status_no_except);
+  result.waitForFinished();
+
+  if (result.result() == 2){
+    long_operation_in_progress = true;
+  }
+  if (result.result() !=0){
+      return;
+  }
 
   initialTimeReset();
 
@@ -1471,7 +1668,7 @@ void MainWindow::on_DeviceConnected() {
           data["initiated_ask"] = !false; //FIXME select proper variable s.NewSDCardFound_u8
           data["erased"] = !s.NewSDCardFound_u8;
           data["erased_ask"] = !s.SDFillWithRandomChars_u8;
-          data["v0.48"] = s.versionInfo.major == 0 && s.versionInfo.minor == 48;
+          data["old_firmware"] = s.versionInfo.major == 0 && s.versionInfo.minor <= 49;
         }
         data["PWS_Access"] = libada::i()->isPasswordSafeUnlocked();
       }
@@ -1487,26 +1684,41 @@ void MainWindow::on_DeviceConnected() {
       if(!data["storage_connected"].toBool()) return;
 
       if (!data["initiated"].toBool()) {
-        if (data["initiated_ask"].toBool())
-          csApplet()->warningBox(tr("Warning: Encrypted volume is not secure,\nSelect \"Initialize "
-                                        "device\" option from context menu."));
+        if (data["initiated_ask"].toBool()){
+          csApplet()->warningBox(tr(Warning_ev_not_secure_initialize) + " " + "\n" + tr(Tray_location_msg));
+          ui->statusBar->showMessage(tr(Warning_ev_not_secure_initialize));
+          make_UI_enabled(false);
+        }
       }
       if (data["initiated"].toBool() && !data["erased"].toBool()) {
         if (data["erased_ask"].toBool())
           csApplet()->warningBox(tr("Warning: Encrypted volume is not secure,\nSelect \"Initialize "
-                                        "storage with random data\""));
+                                        "storage with random data\"") + ". " + "\n" + tr(Tray_location_msg));
       }
 
 #if defined(Q_OS_MAC) || defined(Q_OS_DARWIN)
-      if(data["v0.48"].toBool()){
-        csApplet()->warningBox(tr("Warning: This Storage firmware version (v0.48) is not correctly handled on "
-                                      "MacOS. Application may be unresponsive and unlocking encrypted volume "
-                                      "may not work. Please update the firmware to v0.47 "
-                                      "or use the other latest one when available."));
+      if(data["old_firmware"].toBool()){
+        csApplet()->warningBox(tr(
+                                  "WARNING: This Storage firmware version is old. Application may be unresponsive and unlocking encrypted volume may not work. Please update the firmware to the latest version."
+                                   " "
+                                   "Guide should be available at: <br/><a href='https://www.nitrokey.com/en/doc/firmware-update-storage'>www.nitrokey.com/en/doc/firmware-update-storage</a>."
+                                  ));
       }
 #endif
       }, this);
 
+  QSettings settings;
+  if (settings.value("main/show_main_on_connection", true).toBool()){
+    startConfiguration(false);
+  }
+
+  auto connected_device_model = libada::i()->isStorageDeviceConnected() ?
+                                  tr("Nitrokey Storage connected") :
+                                  tr("Nitrokey Pro connected");
+  if(settings.value("main/connection_message", true).toBool()){
+    tray.showTrayMessage(tr("Nitrokey connected"), connected_device_model);
+  }
+  ui->statusBar->showMessage(connected_device_model);
 }
 
 void MainWindow::on_KeepDeviceOnline() {
@@ -1596,3 +1808,72 @@ void MainWindow::set_debug_mode() {
 void MainWindow::set_debug_level(int debug_level) {
   nm::instance()->set_loglevel(debug_level);
 }
+
+void MainWindow::on_btn_writeSettings_clicked()
+{
+    QSettings settings;
+
+    // see if restart is required
+    bool restart_required = false;
+    if (settings.value("main/language").toString() != ui->combo_languages->currentData().toString()
+            || settings.value("debug/enabled").toBool() != ui->cb_debug_enabled->isChecked()
+            || settings.value("debug/file").toString() != ui->edit_debug_file_path->text()
+            ){
+        restart_required = true;
+    }
+
+    // save the settings
+    settings.setValue("main/first_run", ui->cb_first_run_message->isChecked());
+    settings.setValue("main/language", ui->combo_languages->currentData());
+    settings.setValue("debug/file", ui->edit_debug_file_path->text());
+    settings.setValue("debug/level", ui->spin_debug_verbosity->text().toInt());
+    settings.setValue("debug/enabled", ui->cb_debug_enabled->isChecked());
+    settings.setValue("clipboard/PWS_time", ui->spin_PWS_time->value());
+    settings.setValue("clipboard/OTP_time", ui->spin_OTP_time->value());
+    settings.setValue("main/connection_message", ui->cb_device_connection_message->isChecked());
+    settings.setValue("main/show_main_on_connection", ui->cb_show_main_window_on_connection->isChecked());
+    settings.setValue("main/close_main_on_connection", ui->cb_hide_main_window_on_connection->isChecked());
+    settings.setValue("main/hide_on_close", ui->cb_hide_main_window_on_close->isChecked());
+
+    settings.setValue("storage/check_symlink", ui->cb_check_symlink->isChecked());
+
+    // inform user and quit if asked
+    if (!restart_required){
+        csApplet()->messageBox(tr("Configuration successfully written."));
+    } else {
+        auto user_wants_quit = csApplet()->yesOrNoBox(
+                tr("Configuration successfully written.") + " "+
+                tr("Please run the application again to apply new settings.") + " "+
+                tr("Would you like to quit now?"), true);
+        if (user_wants_quit)
+        {
+            QApplication::exit();
+        }
+    }
+  load_settings_page();
+}
+
+void MainWindow::on_btn_select_debug_file_path_clicked()
+{
+    auto filename = QFileDialog::getSaveFileName(this, tr("Debug file location (will be overwritten)"));
+    ui->edit_debug_file_path->setText(filename);
+}
+
+void MainWindow::on_PWS_Lock_clicked()
+{
+  startLockDeviceAction(true);
+}
+
+void MainWindow::on_btn_copyToClipboard_clicked()
+{
+    clipboard.copyOTP(ui->secretEdit->text());
+    showNotificationLabel();
+}
+
+void MainWindow::ready() {
+}
+
+void MainWindow::on_btn_select_debug_console_clicked()
+{
+    ui->edit_debug_file_path->setText("console");
+}
